
/*
    nc - partially compresses nes files (all compressed parts must have
         a total uncompressed size of 128K or less)
    Copyright (C) 2005 Kuwanger

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    
    History
    v1.0 - September 18, 2005 - Based off gc, which was based off generic_cmp
                              - Now uses mmap, handles args, and tries to do
			        at least some sanity checks (no recompression)
			      - Only supports lzo compression, for the moment
			      - Extension npc stands for nes partial compression
    v1.1 - September 21, 2005 - Includes checksum in header at offset 0x8
    v1.2 - September 22, 2005 - Tries to optimize space more
    v1.3 - September 24, 2005 - Make to work under Windows
    */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <lzo1x.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#define SI STDIN_FILENO
#define SO STDOUT_FILENO

#define LZO_COMPRESSION 0

struct header {
	unsigned int size;
	unsigned int type; 
};

#define HEADER (sizeof(struct header))
#define SIZE (128*1024)
#define OVERHEAD (SIZE/64+16+4+HEADER)
#define WORSTCASE (SIZE+OVERHEAD)

#ifndef O_BINARY
#define O_BINARY 0
#endif

typedef unsigned char u8;
typedef unsigned int u32;

//quick & dirty rom checksum
u32 checksum(u8 *p) {
	u32 sum=0;
	int i;
	for(i=0;i<128;i++) {
		sum+=*p|(*(p+1)<<8)|(*(p+2)<<16)|(*(p+3)<<24);
		p+=128;
	}
	return sum;
}

int usage(char *name)
{
	printf("Usage: %s <file.nes> <file.npc>\n", name);
	return 1;
}

int
main (int argc, char *argv[])
{
  struct header *h;
  struct stat buf;
  char *src, *prg, *chr, *workspace, *temp;
  char *prgcmp, *chrcmp;
  int prgcmp_size, chrcmp_size;
  int in, out, prg_size, chr_size, free_space;
  
  workspace=malloc(448*1024);
  if (!workspace) {
	printf("Out of memory.\n");
	return 2;
  }

  lzo_init ();

  prgcmp = chrcmp = NULL;
  
  if (argc == 3) {
	in = open(argv[1], O_RDONLY | O_BINARY);
	if (in < 0)
		return usage(argv[0]);
	fstat(in, &buf);
	src = malloc(buf.st_size);
	if (!src) {
		printf("Out of memory.\n");
		return 2;
	}
	if (buf.st_size != read(in, src, buf.st_size)) {
		printf("Error reading nes file.\n");
		return 3;
	}
	close(in);
	if (src[4]&0x80 || src[5]&0x80)
		return usage(argv[0]);
	prg_size = src[4]*16384;
	chr_size = src[5]*8192;
	prg = &src[16];
	chr = &src[16+prg_size];
	*((u32 *) &src[8]) = checksum(prg);
	free_space = SIZE;
	if (prg_size && prg_size <= SIZE) {
  		temp = malloc(WORSTCASE);
		if (!temp) {
			printf("Out of memory.\n");
			return 2;
		}
		lzo1x_999_compress(prg, prg_size, temp+HEADER, &prgcmp_size, workspace);
		prgcmp_size += HEADER;
		prgcmp_size = (prgcmp_size+3)&~3;
		h = (struct header *) temp;
		h->size = prgcmp_size;
		h->type = LZO_COMPRESSION;
		prgcmp = temp;
	}
	if (chr_size && chr_size <= SIZE) {
		temp = malloc(WORSTCASE);
		if (!temp) {
			printf("Out of memory.\n");
			return 2;
		}
		lzo1x_999_compress(chr, chr_size, temp+HEADER, &chrcmp_size, workspace);
		chrcmp_size += HEADER;
		chrcmp_size = (chrcmp_size+3)&~3;
		h = (struct header *) temp;
		h->size = chrcmp_size;
		h->type = LZO_COMPRESSION;
		chrcmp = temp;
	}
	if (prgcmp || chrcmp) {
		if (prgcmp && chrcmp && prg_size + chr_size > SIZE) {
			if (prg_size - prgcmp_size > chr_size - chrcmp_size) {
				prg = prgcmp;
				prg_size = prgcmp_size;
				src[4] |= 0x80;
			} else {
				chr = chrcmp;
				chr_size = chrcmp_size;
				src[5] |= 0x80;
			}
		} else {
			if (prgcmp) {
				prg = prgcmp;
				prg_size = prgcmp_size;
				src[4] |= 0x80;
			}
			if (chrcmp) {
				chr = chrcmp;
				chr_size = chrcmp_size;
				src[5] |= 0x80;
			}
		}
		//src[4]&0x80 || src[5]&0x80) {
		out = open(argv[2], O_CREAT | O_WRONLY | O_TRUNC | O_BINARY, 0600);
		if (out < 0)
			return usage(argv[0]);
		write(out, src, 16);
		write(out, prg, prg_size);
		write(out, chr, chr_size);
		close(out);
		printf("Outsize: %d\n", prg_size+chr_size+16);
	} else
		printf("Unable to compress.\n");
	if (prgcmp)
		free(prgcmp);
	if (chrcmp)
		free(chrcmp);
	free(src);
	free(workspace);
  } else {
	free(workspace);
	return usage(argv[0]);
  }
  return 0; 
}

